/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/

/**********************************************************************

grid_proxy_init.c

Description:
	Program used to get a session "proxy" certificate using
	your long term certificate. This is functionally equivelent 
	to the Kerberos kinit program. 

CVS Information:

	$Source: /cvs/nordugrid/grid-manager/external/voms-1.1.39/Socket++/newca.c,v $
	$Date: 2003/11/20 21:12:59 $
	$Revision: 1.3 $
	$Author: aleks $

**********************************************************************/

static char *rcsid = "$Header: /cvs/nordugrid/grid-manager/external/voms-1.1.39/Socket++/newca.c,v 1.3 2003/11/20 21:12:59 aleks Exp $";

/**********************************************************************
                             Include header files
**********************************************************************/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef WIN32
#include "winglue.h"
#include <io.h>
#else
#include <dirent.h>
#endif

/* #include "sslutils.h" */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <sys/stat.h>

#include <openssl/buffer.h>
#include <openssl/crypto.h>
#include <openssl/objects.h>
#include <openssl/asn1.h>
#include <openssl/evp.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/rsa.h>
#include <openssl/lhash.h>
#include <openssl/rand.h>

#ifdef USE_PKCS11
#include "scutils.h"
#endif

#if SSLEAY_VERSION_NUMBER >= 0x0090581fL
#include "openssl/x509v3.h"
#endif

#include "gssapi_compat.h"
#include <gssapi.h>
#include <globus_gsi_cert_utils.h>

#include "globus_oldgaa.h"
#include "globus_oldgaa_utils.h"

#define CANAMESIZE 2048

extern char caname[CANAMESIZE];

#ifdef WIN32
/*********************************************************************
Function: getuid, getpid

Descriptions:
        For Windows95, WIN32, we don't have these, so we will default
    to using uid 0 and pid 0 Need to look at this better for NT. 
******************************************************************/
static unsigned long
getuid()
{
    return 0;
}

static int
getpid()
{
    return 0;
}

#endif /* WIN32 */

static int get_issuer(X509 *cert, char **buffer)
{
  X509_NAME *name;
  char *result;
  int gotit = 0;

  name = X509_get_issuer_name(cert);
  result = X509_NAME_oneline(name, NULL, 0);

  free(*buffer);
  *buffer = (char *)malloc(strlen(result)+1);
  if (*buffer) {
    strncpy(*buffer, result, strlen(result)+1);
    gotit = 1;
  }
  OPENSSL_free(result);
  return gotit;
}

static int get_CA(STACK_OF(X509) *chain, X509 *c, char **buffer)
{
  X509 *cert;
  int num;
  int i;
  int res;
  globus_gsi_cert_utils_cert_type_t type;

  if (!buffer)
    return 0;



  if(globus_gsi_cert_utils_get_cert_type(c,&type) != GLOBUS_SUCCESS) {
    cert=c;
  } else if(!GLOBUS_GSI_CERT_UTILS_IS_PROXY(type)) {
    cert=c;
  } else if (chain) {
    num = sk_X509_num(chain);

    for (i=0; i< num; i++) {
      cert = sk_X509_value(chain, i);
      if(globus_gsi_cert_utils_get_cert_type(c,&type) != GLOBUS_SUCCESS) break;
      if(!GLOBUS_GSI_CERT_UTILS_IS_PROXY(type)) break;
    }
    if (res)
      cert = c;
  }
  else
    cert = c;

/*
  if (!proxy_check_proxy_name(c))
    cert = c;
  else if (chain) {
    num = sk_X509_num(chain);
  
    for (i=0; i< num; i++) {
      cert = sk_X509_value(chain, i);
      res = proxy_check_proxy_name(cert);
      
      if (res == 0) break;
    }
    if (res)
      cert = c;
  }
  else
    cert = c;
*/
  return get_issuer(cert, buffer);
}

/*
 * The following two functions are an ugly hack of the libglobus_gss*
 * libraries done to let a program discover the CA that issued his own
 * certificate. To do so, you must call the following function. See voms.c
 * for mode of calling.
 */
char *
getMCA(void *credential, int globusver)
{
  static char *buffer = NULL;
  STACK_OF(X509) *stk;
  X509 *cert;

  if (!credential)
    return NULL;

/*
  if (globusver == 20) {
    stk  = ((gss_cred_id_desc *)credential)->pcd->cert_chain;
    cert = ((gss_cred_id_desc *)credential)->pcd->ucert;
  }      
  else
*/
  if((globusver == 22) || (globusver == 24)) {
    stk  = ((gss2_cred_id_desc *)credential)->cred_handle->cert_chain;
    cert = ((gss2_cred_id_desc *)credential)->cred_handle->cert; 
  }
  else
    return NULL;

  if (get_CA(stk, cert, &buffer))
    return buffer;
  else
    return NULL;
}

EVP_PKEY *
get_private_key(void *credential, int globusver)
{
/*  proxy_cred_desc *pcd; */
  globus_gsi_cred_handle_t ggch;
  EVP_PKEY *pkey = NULL;

  if (!credential)
    return NULL;

/*
  if (globusver == 20) {
    pcd = ((gss_cred_id_desc *)credential)->pcd;
    if (pcd)
      pkey = pcd->upkey;
  }
  else
*/
  if ((globusver == 22) || (globusver == 24)) {
    ggch = ((gss2_cred_id_desc *)credential)->cred_handle;
    if (ggch)
      pkey = ggch->key;
  }
  else
    return NULL;

  return pkey;
}

EVP_PKEY *
get_delegated_public_key(gss_ctx_id_t context, int globusver)
{
  EVP_PKEY *pkey = NULL;

  if (!context)
    return NULL;

/*
  if (globusver == 20)
    pkey = X509_extract_key(((((gss_ctx_id_desc *)context)->gs_ssl)->session)->peer);
  else
*/
  if ((globusver == 22) || (globusver == 24))
    pkey = X509_extract_key(((((gss2_ctx_id_desc *)context)->gss_ssl)->session)->peer);

  return pkey;
}

X509 *
get_delegated_cert(gss_ctx_id_t context, int globusver)
{
  X509 *cert = NULL;

  if (!context)
    return NULL;

/*
  if (globusver == 20)
    cert = ((((gss_ctx_id_desc *)context)->gs_ssl)->session)->peer;
  else
*/
  if (globusver == 22)
    cert = ((((gss2_ctx_id_desc *)context)->gss_ssl)->session)->peer;

  return cert;
}

char *
get_peer_CA(gss_ctx_id_t context, int globusver)
{
  static char *buffer = NULL;
  STACK_OF(X509) *stk;
  X509 *cert;

  if (!context)
    return NULL;

/*
  if (globusver == 20)
    return caname;
  else
*/
  if ((globusver == 22) || (globusver == 24)) {
    stk  = ((gss2_ctx_id_desc *)context)->peer_cred_handle->cred_handle->cert_chain;
    cert = ((gss2_ctx_id_desc *)context)->peer_cred_handle->cred_handle->cert;
  }
  else
    return NULL;

  if (get_CA(stk, cert, &buffer))
    return buffer;
  else
    return NULL;
}


#define V1_ROOT (EXFLAG_V1|EXFLAG_SS)
#define ku_reject(x, usage) \
	(((x)->ex_flags & EXFLAG_KUSAGE) && !((x)->ex_kusage & (usage)))
#define xku_reject(x, usage) \
	(((x)->ex_flags & EXFLAG_XKUSAGE) && !((x)->ex_xkusage & (usage)))
#define ns_reject(x, usage) \
	(((x)->ex_flags & EXFLAG_NSCERT) && !((x)->ex_nscert & (usage)))


IMPLEMENT_STACK_OF(X509)
IMPLEMENT_ASN1_SET_OF(X509)

IMPLEMENT_STACK_OF(X509_NAME)

IMPLEMENT_STACK_OF(X509_ATTRIBUTE)
IMPLEMENT_ASN1_SET_OF(X509_ATTRIBUTE)
